@@ -1,12 +1,14 @@
 #!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
+import json
 import locale
 import os
 import shutil
 import sys
 from collections.abc import Generator, Sequence
 from contextlib import contextmanager, suppress
+from gettext import gettext as _
 
 from .borders import load_borders_program
 from .boss import Boss
@@ -42,6 +44,7 @@ from .fast_data_types import (
     glfw_get_monitor_workarea,
     glfw_init,
     glfw_terminate,
+    grab_keyboard,
     is_layer_shell_supported,
     load_png_data,
     mask_kitty_signals_process_wide,
@@ -64,6 +67,7 @@ from .utils import (
     get_custom_window_icon,
     log_error,
     parse_os_window_state,
+    read_shell_environment,
     safe_mtime,
     startup_notification_handler,
 )
@@ -84,9 +88,9 @@ def set_custom_ibeam_cursor() -> None:
         log_error(f'Failed to set custom beam cursor with error: {e}')
 
 
-def load_all_shaders(semi_transparent: bool = False) -> None:
+def load_all_shaders() -> None:
     try:
-        load_shader_programs(semi_transparent)
+        load_shader_programs()
         load_borders_program()
     except CompileError as err:
         raise SystemExit(err)
@@ -188,8 +192,9 @@ def set_cocoa_global_shortcuts(opts: Options) -> dict[str, SingleKey]:
                 func_map[parts].append(single_key)
 
         for ac in ('new_os_window', 'close_os_window', 'close_tab', 'edit_config_file', 'previous_tab',
-                   'next_tab', 'new_tab', 'new_window', 'close_window', 'toggle_macos_secure_keyboard_entry', 'toggle_fullscreen',
-                   'hide_macos_app', 'hide_macos_other_apps', 'minimize_macos_window', 'quit'):
+                   'next_tab', 'new_tab', 'new_window', 'close_window', 'toggle_macos_secure_keyboard_entry',
+                   'toggle_fullscreen', 'macos_cycle_through_os_windows', 'macos_cycle_through_os_windows_backwards',
+                   'hide_macos_app', 'hide_macos_other_apps', 'minimize_macos_window', 'quit', 'search_scrollback'):
             val = get_macos_shortcut_for(func_map, ac)
             if val is not None:
                 global_shortcuts[ac] = val
@@ -205,6 +210,9 @@ def set_cocoa_global_shortcuts(opts: Options) -> dict[str, SingleKey]:
         val = get_macos_shortcut_for(func_map, 'clear_terminal to_cursor_scroll active', lookup_name='clear_screen')
         if val is not None:
             global_shortcuts['clear_screen'] = val
+        val = get_macos_shortcut_for(func_map, 'clear_terminal last_command active', lookup_name='clear_last_command')
+        if val is not None:
+            global_shortcuts['clear_last_command'] = val
         val = get_macos_shortcut_for(func_map, 'load_config_file', lookup_name='reload_config')
         if val is not None:
             global_shortcuts['reload_config'] = val
@@ -221,13 +229,23 @@ def is_panel_kitten() -> bool:
     return _is_panel_kitten
 
 
-def list_monitors() -> None:
+def list_monitors(json_output: bool = False) -> None:
     monitor_names = glfw_get_monitor_names()
     has_descriptions = False
     for (name, desc) in monitor_names:
         if desc:
             has_descriptions = True
             break
+
+    if json_output:
+        if has_descriptions:
+            monitors_list_of_dict = [{'name': name, 'description': desc} for name, desc in monitor_names]
+        else:
+            monitors_list_of_dict = [{'name': name} for name, _ in monitor_names]
+        json.dump(monitors_list_of_dict, sys.stdout, indent=2, sort_keys=True)
+        print()
+        return
+
     isatty = sys.stdout.isatty()
     for (name, desc) in monitor_names:
         if isatty:
@@ -242,8 +260,8 @@ def list_monitors() -> None:
 def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = (), talk_fd: int = -1) -> None:
     global _is_panel_kitten
     _is_panel_kitten = run_app.cached_values_name == 'panel'
-    if _is_panel_kitten and run_app.layer_shell_config and run_app.layer_shell_config.output_name == 'list':
-        list_monitors()
+    if _is_panel_kitten and run_app.layer_shell_config and run_app.layer_shell_config.output_name in ('list', 'listjson'):
+        list_monitors(run_app.layer_shell_config.output_name == 'listjson')
         return
     if is_macos:
         global_shortcuts = set_cocoa_global_shortcuts(opts)
@@ -254,17 +272,26 @@ def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = (),
         global_shortcuts = {}
         set_window_icon()
     if _is_panel_kitten and not is_layer_shell_supported():
-        raise SystemExit('Cannot create panels as the window manager/compositor does not support the neccessary protocols')
+        raise SystemExit('Cannot create panels as the window manager/compositor does not support the necessary protocols')
     pos_x, pos_y = None, None
+    if args.grab_keyboard:
+        grab_keyboard(True)
     with cached_values_for(run_app.cached_values_name) as cached_values:
         if not _is_panel_kitten and not is_wayland():
             if opts.remember_window_position:
-                cached_workarea = cached_values.get('monitor-workarea', ())
+                cached_workarea = tuple(tuple(x) for x in cached_values.get('monitor-workarea', ()))
                 if cached_workarea and glfw_get_monitor_workarea() == tuple(cached_workarea):
                     pos_x, pos_y = cached_values.get('window-pos', (None, None))
             if args.position:
                 pos_x, pos_y = map(int, args.position.lower().partition('x')[::2])
-        startup_sessions = tuple(create_sessions(opts, args, default_session=opts.startup_session))
+        startup_session_error: tuple[Exception, str] | None = None
+        try:
+            startup_sessions = tuple(create_sessions(opts, args, default_session=opts.startup_session))
+        except Exception as e:
+            startup_session_error = (e, (getattr(args, 'session', '') or opts.startup_session or ''))
+            if getattr(args, 'session', ''):
+                args.session = ''
+            startup_sessions = tuple(create_sessions(opts, args))
         wincls = (startup_sessions[0].os_window_class if startup_sessions else '') or args.cls or appname
         winname = (startup_sessions[0].os_window_name if startup_sessions else '') or args.name or wincls or appname
         window_state = (args.start_as if args.start_as and args.start_as != 'normal' else None) or (
@@ -285,6 +312,9 @@ def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = (),
         if bad_lines or boss.misc_config_errors:
             boss.show_bad_config_lines(bad_lines, boss.misc_config_errors)
             boss.misc_config_errors = []
+        if startup_session_error:
+            boss.show_error(_('The startup session was invalid'), _(
+                'Loading the start session file {0} failed, with error:\n{1}').format(startup_session_error[1], startup_session_error[0]))
         try:
             boss.child_monitor.main_loop()
         finally:
@@ -365,10 +395,10 @@ def setup_profiling() -> Generator[None, None, None]:
             print('To view the graphical call data, use: kcachegrind', cg)
 
 
-def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
+def expand_listen_on(listen_on: str, from_config_file: bool, env: dict[str, str]) -> str:
     if from_config_file and listen_on == 'none':
         return ''
-    listen_on = expandvars(listen_on)
+    listen_on = expandvars(listen_on, env)
     if '{kitty_pid}' not in listen_on and from_config_file and listen_on.startswith('unix:'):
         listen_on += '-{kitty_pid}'
     listen_on = listen_on.replace('{kitty_pid}', str(os.getpid()))
@@ -448,8 +478,19 @@ def setup_environment(opts: Options, cli_opts: CLIOptions) -> None:
     if not cli_opts.listen_on:
         cli_opts.listen_on = opts.listen_on
         from_config_file = True
+    if vars := opts.env.pop('read_from_shell', ''):
+        import fnmatch
+        import re
+        senv = read_shell_environment(opts)
+        patterns = tuple(re.compile(fnmatch.translate(x.strip())) for x in vars.split() if x.strip())
+        if patterns:
+            for k, v in senv.items():
+                for pat in patterns:
+                    if pat.match(k) is not None:
+                        opts.env[k] = v
+                        break
     if cli_opts.listen_on:
-        cli_opts.listen_on = expand_listen_on(cli_opts.listen_on, from_config_file)
+        cli_opts.listen_on = expand_listen_on(cli_opts.listen_on, from_config_file, opts.env)
     env = opts.env.copy()
     ensure_kitty_in_path()
     ensure_kitten_in_path()
@@ -518,7 +559,7 @@ def kitty_main(called_from_panel: bool = False) -> None:
     if cli_opts.detach:
         if cli_opts.session == '-':
             from .session import PreReadSession
-            cli_opts.session = PreReadSession(sys.stdin.read(), os.environ)
+            cli_opts.session = PreReadSession(sys.stdin.read(), os.environ, '-', os.path.join(os.getcwd(), '-'))
     if cli_opts.replay_commands:
         from kitty.client import main as client_main
         client_main(cli_opts.replay_commands)
